#include <setjmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>

#define MAX_GROUPS                        6
#define GROUP_SIZE                        50                /* 64 would have been more efficient */
#define MAX_HUFCODE_BITS                  20                /* Longest Huffman code allowed */
#define MAX_SYMBOLS                       (256 + 1 + 1)     /* 256 literals + RUNA + RUNB */
#define SYMBOL_RUNA                       0
#define SYMBOL_RUNB                       1

#define RETVAL_OK                         (0)
#define RETVAL_LAST_BLOCK                 (-1)
#define RETVAL_NOT_BZIP_DATA              (-2)
#define RETVAL_UNEXPECTED_INPUT_EOF       (-3)
#define RETVAL_UNEXPECTED_OUTPUT_EOF      (-4)
#define RETVAL_DATA_ERROR                 (-5)
#define RETVAL_OUT_OF_MEMORY              (-6)
#define RETVAL_OBSOLETE_INPUT             (-7)

#define IOBUF_SIZE                        (4 << 10)

struct group_data
{
 int limit[MAX_HUFCODE_BITS+1];
 int base[MAX_HUFCODE_BITS];
 int permute[MAX_SYMBOLS];
 int minLen, maxLen;
};

typedef struct
{
 int writeCopies,writePos,writeRunCountdown,writeCount,writeCurrent;
 int in_fd,out_fd,inbufCount,inbufPos;
 unsigned char *inbuf;
 unsigned int inbufBitCount, inbufBits;
 unsigned int crc32Table[256],headerCRC, totalCRC, writeCRC;
 unsigned int *dbuf, dbufSize;
 unsigned char selectors[32 << 10];
 struct group_data groups[MAX_GROUPS];
 jmp_buf jmpbuf;
} bunzip_data;

static unsigned int get_bits(bunzip_data *bd, char bits_wanted)
{
 unsigned int bits=0;
 while (bd->inbufBitCount<bits_wanted)
 {
  if(bd->inbufPos==bd->inbufCount) 
  {
   if((bd->inbufCount = read(bd->in_fd, bd->inbuf, IOBUF_SIZE)) <= 0)longjmp(bd->jmpbuf,RETVAL_UNEXPECTED_INPUT_EOF);
   bd->inbufPos=0;
  }
  if(bd->inbufBitCount>=24) 
  {
   bits=bd->inbufBits&((1<<bd->inbufBitCount)-1);
   bits_wanted-=bd->inbufBitCount;
   bits<<=bits_wanted;
   bd->inbufBitCount=0;
  }
  bd->inbufBits=(bd->inbufBits<<8)|bd->inbuf[bd->inbufPos++];
  bd->inbufBitCount+=8;
 }
 bd->inbufBitCount-=bits_wanted;
 bits|=(bd->inbufBits>>bd->inbufBitCount)&((1<<bits_wanted)-1);
 return bits;
}

static int get_next_block(bunzip_data *bd)
{
        struct group_data *hufGroup;
        int dbufCount,nextSym,dbufSize,groupCount,*base,*limit,selector,
                i,j,k,t,runPos,symCount,symTotal,nSelectors,byteCount[256];
        unsigned char uc, symToByte[256], mtfSymbol[256], *selectors;
        unsigned int *dbuf,origPtr;
        dbuf=bd->dbuf;
        dbufSize=bd->dbufSize;
        selectors=bd->selectors;
        i=setjmp(bd->jmpbuf);
        if(i) return i;
        i = get_bits(bd,24);
        j = get_bits(bd,24);
        bd->headerCRC=get_bits(bd,32);
        if ((i == 0x177245) && (j == 0x385090)) return RETVAL_LAST_BLOCK;
        if ((i != 0x314159) || (j != 0x265359)) return RETVAL_NOT_BZIP_DATA;
        if(get_bits(bd,1)) return RETVAL_OBSOLETE_INPUT;
        if((origPtr=get_bits(bd,24)) > dbufSize) return RETVAL_DATA_ERROR;
        t=get_bits(bd, 16);
        symTotal=0;
        for (i=0;i<16;i++) {
                if(t&(1<<(15-i))) {
                        k=get_bits(bd,16);
                        for(j=0;j<16;j++)
                                if(k&(1<<(15-j))) symToByte[symTotal++]=(16*i)+j;
                }
        }
        groupCount=get_bits(bd,3);
        if (groupCount<2 || groupCount>MAX_GROUPS) return RETVAL_DATA_ERROR;
        if(!(nSelectors=get_bits(bd, 15))) return RETVAL_DATA_ERROR;
        for(i=0; i<groupCount; i++) mtfSymbol[i] = i;
        for(i=0; i<nSelectors; i++) {
                for(j=0;get_bits(bd,1);j++) if (j>=groupCount) return RETVAL_DATA_ERROR;
                uc = mtfSymbol[j];
                for(;j;j--) mtfSymbol[j] = mtfSymbol[j-1];
                mtfSymbol[0]=selectors[i]=uc;
        }
        symCount=symTotal+2;
        for (j=0; j<groupCount; j++) {
                unsigned char length[MAX_SYMBOLS],temp[MAX_HUFCODE_BITS+1];
                int        minLen,        maxLen, pp;
                t=get_bits(bd, 5)-1;
                for (i = 0; i < symCount; i++) {
                        for(;;) {
                                if (((unsigned)t) > (MAX_HUFCODE_BITS-1))
                                        return RETVAL_DATA_ERROR;
                                k = get_bits(bd,2);
                                if (k < 2) {
                                        bd->inbufBitCount++;
                                        break;
                                }
                                t+=(((k+1)&2)-1);
                        }
                        length[i]=t+1;
                }
                minLen=maxLen=length[0];
                for(i = 1; i < symCount; i++) {
                        if(length[i] > maxLen) maxLen = length[i];
                        else if(length[i] < minLen) minLen = length[i];
                }
                hufGroup=bd->groups+j;
                hufGroup->minLen = minLen;
                hufGroup->maxLen = maxLen;
                base=hufGroup->base-1;
                limit=hufGroup->limit-1;
                pp=0;
                for(i=minLen;i<=maxLen;i++) {
                        temp[i]=limit[i]=0;
                        for(t=0;t<symCount;t++)
                                if(length[t]==i) hufGroup->permute[pp++] = t;
                }
                for (i=0;i<symCount;i++) temp[length[i]]++;
                pp=t=0;
                for (i=minLen; i<maxLen; i++) {
                        pp+=temp[i];
                        limit[i]= (pp << (maxLen - i)) - 1;
                        pp<<=1;
                        base[i+1]=pp-(t+=temp[i]);
                }
                limit[maxLen+1] = INT_MAX;
                limit[maxLen]=pp+temp[maxLen]-1;
                base[minLen]=0;
        }
        for(i=0;i<256;i++) {
                byteCount[i] = 0;
                mtfSymbol[i]=(unsigned char)i;
        }
        runPos=dbufCount=selector=0;
        for(;;) {
                symCount=GROUP_SIZE-1;
                if(selector>=nSelectors) return RETVAL_DATA_ERROR;
                hufGroup=bd->groups+selectors[selector++];
                base=hufGroup->base-1;
                limit=hufGroup->limit-1;
continue_this_group:
                while (bd->inbufBitCount<hufGroup->maxLen) {
                        if(bd->inbufPos==bd->inbufCount) {
                                j = get_bits(bd,hufGroup->maxLen);
                                goto got_huff_bits;
                        }
                        bd->inbufBits=(bd->inbufBits<<8)|bd->inbuf[bd->inbufPos++];
                        bd->inbufBitCount+=8;
                };
                bd->inbufBitCount-=hufGroup->maxLen;
                j = (bd->inbufBits>>bd->inbufBitCount)&((1<<hufGroup->maxLen)-1);
got_huff_bits:
                i=hufGroup->minLen;
                while(j>limit[i]) ++i;
                bd->inbufBitCount += (hufGroup->maxLen - i);
                if ((i > hufGroup->maxLen)
                        || (((unsigned)(j=(j>>(hufGroup->maxLen-i))-base[i]))
                                >= MAX_SYMBOLS))
                        return RETVAL_DATA_ERROR;
                nextSym = hufGroup->permute[j];
                if (((unsigned)nextSym) <= SYMBOL_RUNB) {
                        if(!runPos) {
                                runPos = 1;
                                t = 0;
                        }
                        t += (runPos << nextSym);
                        runPos <<= 1;
                        goto end_of_huffman_loop;
                }
                if(runPos) {
                        runPos=0;
                        if(dbufCount+t>=dbufSize) return RETVAL_DATA_ERROR;

                        uc = symToByte[mtfSymbol[0]];
                        byteCount[uc] += t;
                        while(t--) dbuf[dbufCount++]=uc;
                }
                if(nextSym>symTotal) break;
                if(dbufCount>=dbufSize) return RETVAL_DATA_ERROR;
                i = nextSym - 1;
                uc = mtfSymbol[i];
                do {
                        mtfSymbol[i] = mtfSymbol[i-1];
                } while (--i);
                mtfSymbol[0] = uc;
                uc=symToByte[uc];
                byteCount[uc]++;
                dbuf[dbufCount++] = (unsigned int)uc;
end_of_huffman_loop:
                if(symCount--) goto continue_this_group;
        }
        j=0;
        for(i=0;i<256;i++) {
                k=j+byteCount[i];
                byteCount[i] = j;
                j=k;
        }
        for (i=0;i<dbufCount;i++) {
                uc=(unsigned char)(dbuf[i] & 0xff);
                dbuf[byteCount[uc]] |= (i << 8);
                byteCount[uc]++;
        }
        if(dbufCount) {
                if(origPtr>=dbufCount) return RETVAL_DATA_ERROR;
                bd->writePos=dbuf[origPtr];
            bd->writeCurrent=(unsigned char)(bd->writePos&0xff);
                bd->writePos>>=8;
                bd->writeRunCountdown=5;
        }
        bd->writeCount=dbufCount;
        return RETVAL_OK;
}

static int read_bunzip(bunzip_data *bd, char *outbuf, int len)
{
        const unsigned int *dbuf;
        int pos,current,previous,gotcount;
        if(bd->writeCount<0) return bd->writeCount;
        gotcount = 0;
        dbuf=bd->dbuf;
        pos=bd->writePos;
        current=bd->writeCurrent;
        if (bd->writeCopies) {
                --bd->writeCopies;
                for(;;) {
                        if(gotcount >= len) {
                                bd->writePos=pos;
                                bd->writeCurrent=current;
                                bd->writeCopies++;
                                return len;
                        }
                        outbuf[gotcount++] = current;
                        bd->writeCRC=(((bd->writeCRC)<<8) ^bd->crc32Table[((bd->writeCRC)>>24)^current]);
                        if (bd->writeCopies) {
                                --bd->writeCopies;
                                continue;
                        }
decode_next_byte:
                        if (!bd->writeCount--) break;
                        previous=current;
                        pos=dbuf[pos];
                        current=pos&0xff;
                        pos>>=8;
                        if(--bd->writeRunCountdown) {
                                if(current!=previous) bd->writeRunCountdown=4;
                        } else {
                                bd->writeCopies=current;
                                current=previous;
                                bd->writeRunCountdown=5;
                                if(!bd->writeCopies) goto decode_next_byte;
                                --bd->writeCopies;
                        }
                }
                bd->writeCRC=~bd->writeCRC;
                bd->totalCRC=((bd->totalCRC<<1) | (bd->totalCRC>>31)) ^ bd->writeCRC;
                if(bd->writeCRC!=bd->headerCRC) {
                        bd->totalCRC=bd->headerCRC+1;
                        return RETVAL_LAST_BLOCK;
                }
        }
        previous=get_next_block(bd);
        if(previous) {
                bd->writeCount=previous;
                return (previous!=RETVAL_LAST_BLOCK) ? previous : gotcount;
        }
        bd->writeCRC=0xffffffffUL;
        pos=bd->writePos;
        current=bd->writeCurrent;
        goto decode_next_byte;
}

static int start_bunzip(bunzip_data **bdp, int in_fd, char *inbuf, int len)
{
 bunzip_data *bd;
 unsigned int i,j,c;
 const unsigned int BZh0=(((unsigned int)'B')<<24)+(((unsigned int)'Z')<<16) +(((unsigned int)'h')<<8)+(unsigned int)'0';
 i=sizeof(bunzip_data);
 if(in_fd!=-1) i+=IOBUF_SIZE;
 bd=*bdp=malloc(i);
 memset(bd,0,sizeof(bunzip_data));
 if(-1==(bd->in_fd=in_fd)) {
         bd->inbuf=(unsigned char *)inbuf;
         bd->inbufCount=len;
 } else bd->inbuf=(unsigned char *)(bd+1);
 for(i=0;i<256;i++) {
         c=i<<24;
         for(j=8;j;j--)
                 c=c&0x80000000 ? (c<<1)^0x04c11db7 : (c<<1);
         bd->crc32Table[i]=c;
 }
 i=setjmp(bd->jmpbuf);
 if(i) return i;
 i = get_bits(bd,32);
 if (((unsigned int)(i-BZh0-1)) >= 9) return RETVAL_NOT_BZIP_DATA;
 bd->dbufSize=100000*(i-BZh0);
 bd->dbuf=malloc(bd->dbufSize * sizeof(int));
 return RETVAL_OK;
}

extern int uncompressStream(int src_fd, int dst_fd)
{
 int s_result;
 void *s_buffer;
 bunzip_data *bd;
 s_buffer = malloc(IOBUF_SIZE);
 s_result = start_bunzip(&bd, src_fd, 0, 0);
 if(s_result == 0)
 {
  do
  {
   s_result = read_bunzip(bd, (char *)s_buffer, IOBUF_SIZE);
   if(s_result <= 0)break;
   if(s_result != write(dst_fd, s_buffer, s_result)) 
   {
    s_result = RETVAL_UNEXPECTED_OUTPUT_EOF;
    break;
   }
  }while(1);
 }
 if(s_result == RETVAL_LAST_BLOCK)
 {
  if (bd->headerCRC != bd->totalCRC)(void)fprintf(stderr, "\x1b[1;31mData integrity error when decompressing !\x1b[0m");
  else s_result = RETVAL_OK;
 }
 else if (s_result == RETVAL_UNEXPECTED_OUTPUT_EOF)(void)fprintf(stderr, "\x1b[1;31mCompressed file ends unexpectedly\x1b[0m");
 else (void)fprintf(stderr, "Decompression failed");
 if(bd->dbuf)free(bd->dbuf);
 free(bd);
 free(s_buffer);
 return(s_result);
}

int main(void)
{
 int s_result;
 unsigned char s_byte;
 s_result = uncompressStream(STDIN_FILENO, STDOUT_FILENO);
 if(s_result == 0)
 {
  if(read(STDIN_FILENO, (void *)(&s_byte), sizeof(s_byte)) > 0)(void)fprintf(stderr, "\x1b[1;31mtailing garbage data given !\x1b[0m\n"); 
 }
 else
 {
  static char * const cg_error_message[] = {
   (char *)0,
   "Bad file checksum",
   "Not bzip data",
   "Unexpected input EOF",
   "Unexpected output EOF",
   "Data error",
   "Out of memory",
   "Obsolete (pre 0.9.5) bzip format not supported."
  };
  (void)fprintf(stderr, "bunzip2 error : \"%s\"\n", cg_error_message[-s_result]); 
 }
 
 return(-s_result);
}

/* vim: set expandtab: */
/* End of source */
